from importlib import metadata

from langchain_asteroid.chat_models import ChatAsteroid
from langchain_asteroid.document_loaders import AsteroidLoader
from langchain_asteroid.embeddings import AsteroidEmbeddings
from langchain_asteroid.retrievers import AsteroidRetriever
from langchain_asteroid.toolkits import AsteroidToolkit
from langchain_asteroid.tools import AsteroidTool
from langchain_asteroid.vectorstores import AsteroidVectorStore

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)

__all__ = [
    "ChatAsteroid",
    "AsteroidVectorStore",
    "AsteroidEmbeddings",
    "AsteroidLoader",
    "AsteroidRetriever",
    "AsteroidToolkit",
    "AsteroidTool",
    "__version__",
]
